@@ -60,12 +60,6 @@ fmi3IntervalNotYetKnown = 0
 fmi3IntervalUnchanged   = 1
 fmi3IntervalChanged     = 2
 
-# enum fmi3EventQualifier
-fmi3EventQualifier = c_int
-fmi3EventFalse     = 0
-fmi3EventTrue      = 1
-fmi3EventUnknown   = 2
-
 # callback functions
 fmi3LogMessageCallback         = CFUNCTYPE(None, fmi3InstanceEnvironment, fmi3Status, fmi3String, fmi3String)
 fmi3ClockUpdateCallback        = CFUNCTYPE(None, fmi3InstanceEnvironment)
@@ -169,14 +163,7 @@ class _FMU3(_FMU):
 
         self._fmi3Function('fmi3ExitInitializationMode', [(fmi3Instance, 'instance')])
 
-        self._fmi3Function('fmi3EnterEventMode', [
-            (fmi3Instance,       'instance'),
-            (fmi3EventQualifier, 'stepEvent'),
-            (fmi3EventQualifier, 'stateEvent'),
-            (POINTER(fmi3Int32), 'rootsFound'),
-            (c_size_t,           'nEventIndicators'),
-            (fmi3EventQualifier, 'timeEvent'),
-        ])
+        self._fmi3Function('fmi3EnterEventMode', [(fmi3Instance, 'instance')])
 
         self._fmi3Function('fmi3Terminate', [(fmi3Instance, 'instance')])
 
@@ -281,7 +268,7 @@ class _FMU3(_FMU):
             (c_size_t,          'size')
         ])
 
-        self._fmi3Function('fmi3DeSerializeFMUState', [
+        self._fmi3Function('fmi3DeserializeFMUState', [
             (fmi3Instance,          'instance'),
             (POINTER(fmi3Byte),     'serializedState'),
             (c_size_t,              'size'),
@@ -331,7 +318,7 @@ class _FMU3(_FMU):
             (fmi3Instance,                   'instance'),
             (POINTER(fmi3ValueReference),    'valueReferences'),
             (c_size_t,                       'nValueReferences'),
-            (POINTER(fmi3UInt64),            'intervalCounters'),
+            (POINTER(fmi3UInt64),            'counters'),
             (POINTER(fmi3UInt64),            'resolutions'),
             (POINTER(fmi3IntervalQualifier), 'qualifiers')
         ])
@@ -347,7 +334,7 @@ class _FMU3(_FMU):
             (fmi3Instance,                   'instance'),
             (POINTER(fmi3ValueReference),    'valueReferences'),
             (c_size_t,                       'nValueReferences'),
-            (POINTER(fmi3UInt64),            'shiftCounters'),
+            (POINTER(fmi3UInt64),            'counters'),
             (POINTER(fmi3UInt64),            'resolutions')
         ])
 
@@ -362,7 +349,7 @@ class _FMU3(_FMU):
             (fmi3Instance,                'instance'),
             (POINTER(fmi3ValueReference), 'valueReferences'),
             (c_size_t,                    'nValueReferences'),
-            (POINTER(fmi3UInt64),         'intervalCounters'),
+            (POINTER(fmi3UInt64),         'counters'),
             (POINTER(fmi3UInt64),         'resolutions')
         ])
 
@@ -377,7 +364,7 @@ class _FMU3(_FMU):
             (fmi3Instance,                'instance'),
             (POINTER(fmi3ValueReference), 'valueReferences'),
             (c_size_t,                    'nValueReferences'),
-            (POINTER(fmi3UInt64),         'shiftCounters'),
+            (POINTER(fmi3UInt64),         'counters'),
             (POINTER(fmi3UInt64),         'resolutions')
         ])
 
@@ -591,18 +578,8 @@ class _FMU3(_FMU):
     def setIntervalFraction(self, valueReferences, intervalCounters, resolutions):
         self.fmi3SetIntervalFraction(self.component, valueReferences, len(valueReferences), intervalCounters, resolutions, len(intervalCounters))
 
-    def enterEventMode(self, stepEvent=False, stateEvent=False, rootsFound=[], timeEvent=False):
-
-        rootsFound = (fmi3Int32 * len(rootsFound))(*rootsFound)
-
-        return self.fmi3EnterEventMode(
-            self.component,
-            fmi3EventQualifier(stepEvent),
-            fmi3EventQualifier(stateEvent),
-            rootsFound,
-            len(rootsFound),
-            fmi3EventQualifier(timeEvent)
-        )
+    def enterEventMode(self):
+        self.fmi3EnterEventMode(self.component)
 
     def updateDiscreteStates(self):
 
@@ -850,16 +827,18 @@ class _FMU3(_FMU):
         self.fmi3SerializeFMUState(self.component, state, serializedState, size)
         return serializedState.raw
 
-    def deSerializeFMUState(self, serializedState, state):
+    def deserializeFMUState(self, serializedState, state=None):
         """ De-serialize an FMU state
 
         Parameters:
             serializedState   the serialized state as a byte string
             state             the FMU state
         """
-
-        buffer = create_string_buffer(serializedState)
-        self.fmi3DeSerializeFMUState(self.component, buffer, len(buffer), byref(state))
+        if state is None:
+            state = fmi3FMUState()
+        buffer = create_string_buffer(serializedState, size=len(serializedState))
+        self.fmi3DeserializeFMUState(self.component, buffer, len(buffer), byref(state))
+        return state
 
     # Getting partial derivatives
 
