@@ -112,7 +112,19 @@ class Recorder(object):
         if modelDescription.fmiVersion in ['1.0', '2.0']:
             types = [('Real', np.float64), ('Integer', np.int32), ('Boolean', np.bool_)]
         else:
-            types = [('Float64', np.float64), ('Int32', np.int32), ('UInt64', np.uint64), ('Boolean', np.bool_)]
+            types = [
+                ('Float32', np.float32),
+                ('Float64', np.float64),
+                ('Int8', np.int8),
+                ('UInt8', np.uint8),
+                ('Int16', np.int16),
+                ('UInt16', np.uint16),
+                ('Int32', np.int32),
+                ('UInt32', np.uint32),
+                ('Int64', np.int64),
+                ('UInt64', np.uint64),
+                ('Boolean', np.bool_),
+            ]
 
         # collect the columns
         for t, dt in types:
@@ -226,8 +238,16 @@ class Input(object):
             setters['Integer'] = (fmu.fmi2SetInteger, fmi2Integer)
             setters['Boolean'] = (fmu.fmi2SetBoolean, fmi2Boolean)
         else:
+            setters['Float32'] = (fmu.fmi3SetFloat32, fmi3.fmi3Float32)
             setters['Float64'] = (fmu.fmi3SetFloat64, fmi3.fmi3Float64)
+            setters['Int8']    = (fmu.fmi3SetInt8,    fmi3.fmi3Int8)
+            setters['UInt8']   = (fmu.fmi3SetUInt8,   fmi3.fmi3UInt8)
+            setters['Int16']   = (fmu.fmi3SetInt16,   fmi3.fmi3Int16)
+            setters['UInt16']  = (fmu.fmi3SetUInt16,  fmi3.fmi3UInt16)
             setters['Int32']   = (fmu.fmi3SetInt32,   fmi3.fmi3Int32)
+            setters['UInt32']  = (fmu.fmi3SetUInt32,  fmi3.fmi3UInt32)
+            setters['Int64']   = (fmu.fmi3SetInt64,   fmi3.fmi3Int64)
+            setters['UInt64']  = (fmu.fmi3SetUInt64,  fmi3.fmi3UInt64)
             setters['Boolean'] = (fmu.fmi3SetBoolean, fmi3.fmi3Boolean)
 
         from collections import defaultdict
@@ -289,6 +309,7 @@ class Input(object):
         if self.t is None:
             return
 
+        is_fmi1 = isinstance(self.fmu, _FMU1)
         is_fmi3 = isinstance(self.fmu, fmi3._FMU3)
 
         # continuous
@@ -308,7 +329,7 @@ class Input(object):
             for vrs, values, table, setter in self.discrete:
                 values[:], der_values = self.interpolate(time=time, t=self.t, table=table, discrete=True, after_event=after_event)
 
-                if values._type_ == c_int8:
+                if is_fmi1 and values._type_ == c_int8:
                     # special treatment for fmi1Boolean
                     setter(self.fmu.component, vrs, len(vrs), cast(values, POINTER(c_char)))
                 else:
@@ -566,7 +587,7 @@ def simulate_fmu(filename,
                  validate: bool = True,
                  start_time: Union[float, str] = None,
                  stop_time: Union[float, str] = None,
-                 solver: str ='CVode',
+                 solver: str = 'CVode',
                  step_size: Union[float, str] = None,
                  relative_tolerance: Union[float, str] = None,
                  output_interval: Union[float, str] = None,
@@ -587,7 +608,10 @@ def simulate_fmu(filename,
                  set_input_derivatives: bool = False,
                  remote_platform: str = 'auto',
                  early_return_allowed: bool = False,
-                 use_event_mode: bool = False) -> SimulationResult:
+                 use_event_mode: bool = False,
+                 initialize: bool = True,
+                 terminate: bool = True,
+                 fmu_state: Union[bytes, c_void_p] = None) -> SimulationResult:
     """ Simulate an FMU
 
     Parameters:
@@ -618,6 +642,9 @@ def simulate_fmu(filename,
                                is not supported, None: no remoting; experimental)
         early_return_allowed   allow early return in FMI 3.0 Co-Simulation
         use_event_mode         use event mode in FMI 3.0 Co-Simulation if the FMU supports it
+        initialize             initialize the FMU
+        terminate              terminate the FMU
+        fmu_state              the FMU state or serialized FMU state to initialize the FMU
     Returns:
         result                 a structured numpy array that contains the result
     """
@@ -650,6 +677,12 @@ def simulate_fmu(filename,
     if fmi_type not in ['ModelExchange', 'CoSimulation']:
         raise Exception('fmi_type must be one of "ModelExchange" or "CoSimulation"')
 
+    if initialize is False:
+        if fmi_type != 'CoSimulation':
+            raise Exception("If initialize is False, the interface type must be 'CoSimulation'.")
+        if fmu_instance is None and fmu_state is None:
+            raise Exception("If initialize is False, fmu_instance or fmu_state must be provided.")
+
     experiment = model_description.defaultExperiment
 
     if start_time is None:
@@ -706,11 +739,23 @@ def simulate_fmu(filename,
     else:
         fmu = fmu_instance
 
+    if fmu_state is not None:
+        if model_description.fmiVersion == '2.0' or model_description.fmiVersion.startswith('3.0'):
+            if isinstance(fmu_state, bytes):
+                fmu_state = fmu.deserializeFMUState(fmu_state)
+                fmu.setFMUState(fmu_state)
+                fmu.freeFMUState(fmu_state)
+            else:
+                fmu.setFMUState(fmu_state)
+        else:
+            raise Exception(f"Setting the FMU state is not supported for FMI version {model_description.fmiVersion}.")
+        initialize = False
+
     # simulate_fmu the FMU
     if fmi_type == 'ModelExchange':
         result = simulateME(model_description, fmu, start_time, stop_time, solver, step_size, relative_tolerance, start_values, apply_default_start_values, input, output, output_interval, record_events, timeout, step_finished, validate)
     elif fmi_type == 'CoSimulation':
-        result = simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance, start_values, apply_default_start_values, input, output, output_interval, timeout, step_finished, set_input_derivatives, use_event_mode, early_return_allowed, validate)
+        result = simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance, start_values, apply_default_start_values, input, output, output_interval, timeout, step_finished, set_input_derivatives, use_event_mode, early_return_allowed, validate, initialize, terminate)
 
     if fmu_instance is None:
         fmu.freeInstance()
@@ -738,33 +783,29 @@ def instantiate_fmu(unzipdir, model_description, fmi_type=None, visible=False, d
     if library_path:
         fmu_args['libraryPath'] = library_path
 
-    if logger is None:
-        logger = printLogMessage
-
     is_fmi1 = model_description.fmiVersion == '1.0'
     is_fmi2 = model_description.fmiVersion == '2.0'
 
-    if is_fmi1:
-        callbacks = fmi1CallbackFunctions()
-        callbacks.logger         = fmi1CallbackLoggerTYPE(logger)
-        callbacks.allocateMemory = fmi1CallbackAllocateMemoryTYPE(allocateMemory)
-        callbacks.freeMemory     = fmi1CallbackFreeMemoryTYPE(freeMemory)
-        callbacks.stepFinished   = None
-    elif is_fmi2:
-        callbacks = fmi2CallbackFunctions()
-        callbacks.logger         = fmi2CallbackLoggerTYPE(logger)
-        callbacks.allocateMemory = fmi2CallbackAllocateMemoryTYPE(allocateMemory)
-        callbacks.freeMemory     = fmi2CallbackFreeMemoryTYPE(freeMemory)
-    else:
-        callbacks = None
+    if logger is not None:
+        if is_fmi1:
+            callbacks = fmi1CallbackFunctions()
+            callbacks.logger         = fmi1CallbackLoggerTYPE(logger)
+            callbacks.allocateMemory = fmi1CallbackAllocateMemoryTYPE(calloc)
+            callbacks.freeMemory     = fmi1CallbackFreeMemoryTYPE(free)
+            callbacks.stepFinished   = None
+        elif is_fmi2:
+            callbacks = fmi2CallbackFunctions()
+            callbacks.logger         = fmi2CallbackLoggerTYPE(logger)
+            callbacks.allocateMemory = fmi2CallbackAllocateMemoryTYPE(calloc)
+            callbacks.freeMemory     = fmi2CallbackFreeMemoryTYPE(free)
 
-    if model_description.fmiVersion in ['1.0', '2.0']:
-        # add native proxy function that processes variadic arguments
         try:
             from .logging import addLoggerProxy
             addLoggerProxy(byref(callbacks))
         except Exception as e:
             print("Failed to add logger proxy function. %s" % e)
+    else:
+        callbacks = None
 
     if fmi_type in [None, 'CoSimulation'] and model_description.coSimulation is not None:
 
@@ -778,7 +819,8 @@ def instantiate_fmu(unzipdir, model_description, fmi_type=None, visible=False, d
             fmu.instantiate(visible=visible, callbacks=callbacks, loggingOn=debug_logging)
         else:
             fmu = fmi3.FMU3Slave(**fmu_args)
-            fmu.instantiate(visible=visible, loggingOn=debug_logging, eventModeUsed=event_mode_used, earlyReturnAllowed=early_return_allowed)
+            fmu.instantiate(visible=visible, loggingOn=debug_logging, eventModeUsed=event_mode_used,
+                            earlyReturnAllowed=early_return_allowed, logMessage=logger)
 
     elif fmi_type in [None, 'ModelExchange'] and model_description.modelExchange is not None:
 
@@ -792,13 +834,13 @@ def instantiate_fmu(unzipdir, model_description, fmi_type=None, visible=False, d
             fmu.instantiate(visible=visible, callbacks=callbacks, loggingOn=debug_logging)
         else:
             fmu = fmi3.FMU3Model(**fmu_args)
-            fmu.instantiate(visible=visible, loggingOn=debug_logging)
+            fmu.instantiate(visible=visible, loggingOn=debug_logging, logMessage=logger)
 
     elif fmi_type in [None, 'ScheduledExecution'] and model_description.scheduledExecution is not None:
 
         fmu_args['modelIdentifier'] = model_description.scheduledExecution.modelIdentifier
         fmu = fmi3.FMU3ScheduledExecution(**fmu_args)
-        fmu.instantiate(visible=visible, loggingOn=debug_logging)
+        fmu.instantiate(visible=visible, loggingOn=debug_logging, logMessage=logger)
 
     else:
 
@@ -864,11 +906,11 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
         (iterationConverged,
          stateValueReferencesChanged,
          stateValuesChanged,
-         terminateSimulation,
+         terminate_simulation,
          nextEventTimeDefined,
          nextEventTime) = fmu.initialize()
 
-        if terminateSimulation:
+        if terminate_simulation:
             raise Exception('Model requested termination during initial event update.')
 
     elif is_fmi2:
@@ -884,17 +926,17 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
         fmu.exitInitializationMode()
 
         newDiscreteStatesNeeded = True
-        terminateSimulation = False
+        terminate_simulation = False
 
-        while newDiscreteStatesNeeded and not terminateSimulation:
+        while newDiscreteStatesNeeded and not terminate_simulation:
             (newDiscreteStatesNeeded,
-             terminateSimulation,
+             terminate_simulation,
              nominalsOfContinuousStatesChanged,
              valuesOfContinuousStatesChanged,
              nextEventTimeDefined,
              nextEventTime) = fmu.newDiscreteStates()
 
-        if terminateSimulation:
+        if terminate_simulation:
             raise Exception('Model requested termination during initial event update.')
 
         fmu.enterContinuousTimeMode()
@@ -912,12 +954,12 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
         fmu.exitInitializationMode()
 
         discreteStatesNeedUpdate = True
-        terminateSimulation = False
+        terminate_simulation = False
 
-        while discreteStatesNeedUpdate and not terminateSimulation:
+        while discreteStatesNeedUpdate and not terminate_simulation:
             # update discrete states
             (discreteStatesNeedUpdate,
-             terminateSimulation,
+             terminate_simulation,
              nominalsOfContinuousStatesChanged,
              valuesOfContinuousStatesChanged,
              nextEventTimeDefined,
@@ -1041,15 +1083,15 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
                 iterationConverged = False
 
                 # update discrete states
-                while not iterationConverged and not terminateSimulation:
+                while not iterationConverged and not terminate_simulation:
                     (iterationConverged,
                      stateValueReferencesChanged,
                      stateValuesChanged,
-                     terminateSimulation,
+                     terminate_simulation,
                      nextEventTimeDefined,
                      nextEventTime) = fmu.eventUpdate()
 
-                if terminateSimulation:
+                if terminate_simulation:
                     break
 
             elif is_fmi2:
@@ -1062,22 +1104,22 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
                 newDiscreteStatesNeeded = True
 
                 # update discrete states
-                while newDiscreteStatesNeeded and not terminateSimulation:
+                while newDiscreteStatesNeeded and not terminate_simulation:
                     (newDiscreteStatesNeeded,
-                     terminateSimulation,
+                     terminate_simulation,
                      nominalsOfContinuousStatesChanged,
                      valuesOfContinuousStatesChanged,
                      nextEventTimeDefined,
                      nextEventTime) = fmu.newDiscreteStates()
 
-                if terminateSimulation:
+                if terminate_simulation:
                     break
 
                 fmu.enterContinuousTimeMode()
 
             else:
 
-                fmu.enterEventMode(stepEvent=step_event, rootsFound=roots_found, timeEvent=time_event)
+                fmu.enterEventMode()
 
                 if input_event:
                     input.apply(time=time, after_event=True)
@@ -1085,15 +1127,15 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
                 newDiscreteStatesNeeded = True
 
                 # update discrete states
-                while newDiscreteStatesNeeded and not terminateSimulation:
+                while newDiscreteStatesNeeded and not terminate_simulation:
                     (newDiscreteStatesNeeded,
-                     terminateSimulation,
+                     terminate_simulation,
                      nominalsOfContinuousStatesChanged,
                      valuesOfContinuousStatesChanged,
                      nextEventTimeDefined,
                      nextEventTime) = fmu.updateDiscreteStates()
 
-                if terminateSimulation:
+                if terminate_simulation:
                     break
 
                 fmu.enterContinuousTimeMode()
@@ -1118,7 +1160,7 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
     return recorder.result()
 
 
-def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance, start_values, apply_default_start_values, input_signals, output, output_interval, timeout, step_finished, set_input_derivatives, use_event_mode, early_return_allowed, validate):
+def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance, start_values, apply_default_start_values, input_signals, output, output_interval, timeout, step_finished, set_input_derivatives, use_event_mode, early_return_allowed, validate, initialize, terminate):
 
     if set_input_derivatives and not model_description.coSimulation.canInterpolateInputs:
         raise Exception("Parameter set_input_derivatives is True but the FMU cannot interpolate inputs.")
@@ -1131,42 +1173,44 @@ def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance
     is_fmi1 = model_description.fmiVersion == '1.0'
     is_fmi2 = model_description.fmiVersion == '2.0'
 
-    if is_fmi2:
-        fmu.setupExperiment(tolerance=relative_tolerance, startTime=start_time, stopTime=stop_time)
-
     input = Input(fmu=fmu, modelDescription=model_description, signals=input_signals, set_input_derivatives=set_input_derivatives)
 
     time = start_time
 
-    # initialize the model
-    if is_fmi1:
-        start_values = apply_start_values(fmu, model_description, start_values, settable=has_start_value)
-        input.apply(time)
-        fmu.initialize(tStart=time, stopTime=stop_time)
-    elif is_fmi2:
-        start_values = apply_start_values(fmu, model_description, start_values, settable=settable_in_instantiated)
-        fmu.enterInitializationMode()
-        start_values = apply_start_values(fmu, model_description, start_values, settable=settable_in_initialization_mode)
-        input.apply(time)
-        fmu.exitInitializationMode()
-    else:
-        start_values = apply_start_values(fmu, model_description, start_values, settable=settable_in_instantiated)
-        fmu.enterInitializationMode(tolerance=relative_tolerance, startTime=start_time, stopTime=stop_time)
-        start_values = apply_start_values(fmu, model_description, start_values, settable=settable_in_initialization_mode)
-        input.apply(time)
-        fmu.exitInitializationMode()
+    if initialize:
 
-        if use_event_mode:
+        if is_fmi2:
+            fmu.setupExperiment(tolerance=relative_tolerance, startTime=time, stopTime=stop_time)
 
-            update_discrete_states = True
+        # initialize the model
+        if is_fmi1:
+            start_values = apply_start_values(fmu, model_description, start_values, settable=has_start_value)
+            input.apply(time)
+            fmu.initialize(tStart=time, stopTime=stop_time)
+        elif is_fmi2:
+            start_values = apply_start_values(fmu, model_description, start_values, settable=settable_in_instantiated)
+            fmu.enterInitializationMode()
+            start_values = apply_start_values(fmu, model_description, start_values, settable=settable_in_initialization_mode)
+            input.apply(time)
+            fmu.exitInitializationMode()
+        else:
+            start_values = apply_start_values(fmu, model_description, start_values, settable=settable_in_instantiated)
+            fmu.enterInitializationMode(tolerance=relative_tolerance, startTime=time, stopTime=stop_time)
+            start_values = apply_start_values(fmu, model_description, start_values, settable=settable_in_initialization_mode)
+            input.apply(time)
+            fmu.exitInitializationMode()
 
-            while update_discrete_states:
-                update_discrete_states, terminate, _, _, _, _ = fmu.updateDiscreteStates()
+            if use_event_mode:
 
-            if terminate:
-                raise Exception("The FMU requested to terminate the simulation during initialization.")
+                update_discrete_states = True
 
-            fmu.enterStepMode()
+                while update_discrete_states:
+                    update_discrete_states, terminate_simulation, _, _, _, _ = fmu.updateDiscreteStates()
+
+                if terminate_simulation:
+                    raise Exception("The FMU requested to terminate the simulation during initialization.")
+
+                fmu.enterStepMode()
 
     if validate and len(start_values) > 0:
         raise Exception("The start values for the following variables could not be set: " +
@@ -1176,9 +1220,7 @@ def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance
 
     n_steps = time / output_interval
 
-    terminate = False
-
-    time = start_time
+    terminate_simulation = False
 
     # simulation loop
     while True:
@@ -1188,7 +1230,7 @@ def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance
         if timeout is not None and (current_time() - sim_start) > timeout:
             break
 
-        if terminate or time >= stop_time:
+        if terminate_simulation or time >= stop_time:
             break
 
         input.apply(time)
@@ -1232,7 +1274,7 @@ def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance
 
             step_size = t_input_event - time if input_event else t_next - time
 
-            event_encountered, terminate, early_return, last_successful_time = fmu.doStep(currentCommunicationPoint=time, communicationStepSize=step_size)
+            event_encountered, terminate_simulation, early_return, last_successful_time = fmu.doStep(currentCommunicationPoint=time, communicationStepSize=step_size)
 
             if early_return and not early_return_allowed:
                 raise Exception("FMU returned early from doStep() but Early Return is not allowed.")
@@ -1253,17 +1295,19 @@ def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance
 
                 update_discrete_states = True
 
-                while update_discrete_states and not terminate:
-                    update_discrete_states, terminate, _, _, _, _ = fmu.updateDiscreteStates()
+                while update_discrete_states and not terminate_simulation:
+                    update_discrete_states, terminate_simulation, _, _, _, _ = fmu.updateDiscreteStates()
 
-                if terminate:
+                if terminate_simulation:
                     break
 
                 fmu.enterStepMode()
 
         if step_finished is not None and not step_finished(time, recorder):
+            recorder.sample(time, force=True)
             break
 
-    fmu.terminate()
+    if terminate:
+        fmu.terminate()
 
     return recorder.result()
